#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/msg.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>

#define TRUE 1
#define FALSE 0

int *numberOfScores;
int msqid;
pid_t pid;
int ordinalNumber;

typedef struct {
	long mtype;	
	int statisticsPid;
} HelpMessage;

typedef struct {
	long mtype;
} CollectBallMessage;

union semun {
	int val;
	struct semid_ds *buf;
	unsigned short *array;
};

HelpMessage fetchHelpMessage() {
	HelpMessage helpMessage;
	helpMessage.statisticsPid = -1;

	printf("Cekam poruku : tip %d\n", ordinalNumber);
	msgrcv(msqid, &helpMessage, sizeof(helpMessage) - sizeof(long), ordinalNumber, IPC_NOWAIT);

	return helpMessage;
}

int didPlayerScore() {
	return rand() % 2;
}

void signal_increase_score(int sig) {
	numberOfScores++;
}

int selectRandomHelper(int ordinalNumber) {
	int randomPlayer;

	while ((randomPlayer = (rand() % 4) + 1) == ordinalNumber);
	return randomPlayer;
}

void signal_call_help(int sig) {
	HelpMessage helpMessage;
	helpMessage.mtype = selectRandomHelper(ordinalNumber);
	helpMessage.statisticsPid = pid;

	printf("Saljem poruku : tip %ld, int %d.\n", helpMessage.mtype, helpMessage.statisticsPid);
	if ((msgsnd(msqid, &helpMessage, sizeof(helpMessage) - sizeof(long), 0)) == -1) {
		perror("unable to send help message");
		exit(1);
	}
}

void notifyBallBoy() {
	CollectBallMessage collectBallMessage;
	collectBallMessage.mtype = 5;

	printf("Saljem poruku : tip %ld\n", collectBallMessage.mtype);
	if ((msgsnd(msqid, &collectBallMessage, sizeof(collectBallMessage) - sizeof(long), 0)) == -1) {
		perror("unable to notify ball boy");
		exit(1);
	}
}

int main(int argc, char const *argv[])
{

	ordinalNumber = argv[1][0] - '0';
	printf("Ordinal number : %d\n", ordinalNumber);

	key_t key = ftok("./igrac.c", 'a');
	int msgflg = IPC_CREAT | 0666;
	int semid = semget(key, 2, IPC_CREAT | IPC_EXCL | 0666);
	msqid = msgget(key, msgflg);

	if (semid == -1) {
		if (errno == EEXIST) {
			semid = semget(key, 2, 0);
		} else {
			perror("unable to obtain semaphore id");
			exit(1);
		}
	} else {
		union semun sem_init;

		sem_init.array = (unsigned short*)malloc(2 * sizeof(unsigned short));
		
		int N;
		printf("Unesite limit lopti na stalku:\n");
		scanf("%d", &N);

		sem_init.array[0] = N;
		sem_init.array[1] = 1; // MUTEX

		semctl(semid, 0, SETALL, sem_init);
	}

	int shmid = shmget(IPC_PRIVATE, sizeof(int), IPC_CREAT | 0666);
	numberOfScores = (int*)shmat(shmid, NULL, 0);

	*numberOfScores = 0;

	pid = fork();
	if (pid == -1) {
		perror("unable to fork process");
		exit(1);
	} else if (pid == 0) {
		srand(getpid());

		signal(SIGCHLD, signal_increase_score);
		signal(SIGINT, signal_call_help);

		while (1) {
			printf("Na kosu rbr %d trenutno ima %d pogodaka.\n", ordinalNumber, *numberOfScores);
			sleep(10);
		}
	} else {
		struct sembuf op_take_ball;

		op_take_ball.sem_num = 0;
		op_take_ball.sem_op = -1;
		op_take_ball.sem_flg = 0;

		struct sembuf op_mutex_ball_dec, op_mutex_ball_inc;

		op_mutex_ball_dec.sem_num = 1;
		op_mutex_ball_dec.sem_op = -1;
		op_mutex_ball_dec.sem_flg = 0;

		op_mutex_ball_inc.sem_num = 1;
		op_mutex_ball_inc.sem_op = 1;
		op_mutex_ball_inc.sem_flg = 0;

		HelpMessage helpMessage;
		int basketballOrdinalPid;
		int numberOfMistakes = 0;

		srand(getpid());
		while (1) {
			// TAKING BALL
			printf("Smanjujem semafor %d iz skupa %d za vrednost %d\n", 1, semid, 1);
			sleep(3);
			semop(semid, &op_mutex_ball_dec, 1);

			printf("Smanjujem semafor %d iz skupa %d za vrednost %d\n", 0, semid, 1);
			sleep(3);
			semop(semid, &op_take_ball, 1);
			sleep(1);

			printf("Povecavam semafor %d iz skupa %d za vrednost %d\n", 1, semid, 1);
			sleep(3);
			semop(semid, &op_mutex_ball_inc, 1);
			// FINISHED TAKING BALL

			helpMessage = fetchHelpMessage();
			if (helpMessage.statisticsPid == -1) {
				basketballOrdinalPid = pid;
			} else {
				printf("Primio sam poziv za pomoc!\n");
				basketballOrdinalPid = helpMessage.statisticsPid;
			}

			sleep(3);
			if (didPlayerScore()) {
				printf("Pogodio sam kos!\n");

				if (helpMessage.statisticsPid == -1) {
					kill(pid, SIGCHLD);	
				} else {
					kill(helpMessage.statisticsPid, SIGCHLD);
				}
			} else {
				printf("Promasio sam kos!\n");

				if (helpMessage.statisticsPid == -1) {
					numberOfMistakes++;
					if (numberOfMistakes == 2) {
						numberOfMistakes = 0;
						
						// call for help
						printf("Saljem poziv za pomoc!\n");
						kill(pid, SIGINT);
					}
				} else {
					// NOT DEFINIED IN TASK
				}
			}

			printf("Obavestavam redara!\n");
			notifyBallBoy();
		}
	}
}